#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <asm/types.h>
#include <linux/hdlc.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/if.h>


void run_board(const int sock, struct ifreq *req, const char *filename)
  {
  __u32 length;
  FILE *fw;

  fw=fopen(filename, "rb");
  if (fw==NULL)
    {
    fprintf(stderr, "Unable to open firmware file: %s\n", strerror(errno));
    exit(1);
    }

  fseek(fw, 0, SEEK_END);
  length=ftell(fw);
  fseek(fw, 0, SEEK_SET);
  req->ifr_data=malloc(length+sizeof(__u32));
  if (req->ifr_data==NULL)
    {
    fprintf(stderr, "Out of memory\n");
    fclose(fw);
    exit(1);
    }

  *(__u32*)req->ifr_data = length;
  if (fread(req->ifr_data+sizeof(__u32), 1, length, fw)!=length)
    {
    fprintf(stderr, "Error reading firmware file: %s\n", strerror(errno));
    fclose(fw);
    exit(1);
    }

  fclose(fw);

  if (ioctl(sock, HDLCRUN, req))
    {
    fprintf(stderr, "Running failed: %s\n", strerror(errno));
    exit(1);
    }

  free(req->ifr_data);
  }



void simple_ioctl(const int sock, struct ifreq *req, const int cmd,
		  const int value)
  {
  req->ifr_ifru.ifru_ivalue=value;
  
  if (ioctl(sock, cmd, req))
    {
    fprintf(stderr, "Ioctl failed: %s\n", strerror(errno));
    exit(1);
    }
  }


void set_clock(const int sock, struct ifreq *req, const char *arg)
  {
  unsigned long speed;
  char test;

  if (!strcmp(arg, "ext"))
    {
    simple_ioctl(sock, req, HDLCSETLINE, LINE_EXT_CLOCK);
    return;
    }
  
  if (sscanf(arg, "%ld%c", &speed, &test)==1)
    if (speed>=LINE_MIN_CLOCK && speed)
      {
      simple_ioctl(sock, req, HDLCSETLINE, speed);
      return;
      }
  fprintf(stderr, "Invalid line speed %s\n", arg);
  exit(1);
  }



void fr_pvc(const int sock, struct ifreq *req, char *arg, const int creat_del)
  {
  int dlci;
  char test;

  if (sscanf(arg, "%u%c", &dlci, &test)==1)
    if (dlci>0 && dlci<1023)
      {
      simple_ioctl(sock, req, HDLCPVC, dlci*creat_del);
      return;
      }

  fprintf(stderr, "Invalid dlci %s\n", arg);
  exit(1);
  }



int main(int argc, char *argv[])
  {
  int sock, arg=1;
  struct ifreq req;

  if (argc <= 1)
    {
    fprintf(stderr, "\nUsage: sethdlc iface command [parameters] ...\n"
	    "\n"
	    "commands:\n"
	    "\trun firmware_file_name\n"
	    "\tclock rate (default = ext)\n"
	    "\tmode [soft] [dce] (hdlc | cisco | ppp | x25 | fr_ansi | fr_ccitt)\n"
	    "\tv35 | rs232 | x21 | t1 | e1 | loopback | noloopback\n"
	    "\t(create | delete) dlci\n"
	    "\n");
    exit(0);
    }
  
  sock=socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
  if (sock<0)
    {
    perror("Unable to create socket");
    exit(1);
    }
  
  strcpy(req.ifr_name, argv[1]); /* Device name */

  while (arg++, argc > arg)
    {
    if (argc > arg+1 && !strcmp(argv[arg], "clock"))
      {
      set_clock(sock, &req, argv[++arg]);
      continue;
      }
    
    if (!strcmp(argv[arg], "v35"))
      {
      simple_ioctl(sock, &req, HDLCSETLINE, LINE_V35);
      continue;
      }
    
    if (!strcmp(argv[arg], "rs232"))
      {
      simple_ioctl(sock, &req, HDLCSETLINE, LINE_RS232);
      continue;
      }
    
    if (!strcmp(argv[arg], "x21"))
      {
      simple_ioctl(sock, &req, HDLCSETLINE, LINE_X21);
      continue;
      }
    
    if (!strcmp(argv[arg], "loopback"))
      {
      simple_ioctl(sock, &req, HDLCSETLINE, LINE_LOOPBACK);
      continue;
      }
    
    if (!strcmp(argv[arg], "noloopback"))
      {
      simple_ioctl(sock, &req, HDLCSETLINE, LINE_NOLOOPBACK);
      continue;
      }
    
    if (!strcmp(argv[arg], "t1"))
      {
      simple_ioctl(sock, &req, HDLCSETLINE, LINE_T1);
      continue;
      }
    
    if (!strcmp(argv[arg], "e1"))
      {
      simple_ioctl(sock, &req, HDLCSETLINE, LINE_E1);
      continue;
      }
    
    if (argc > arg+1 && !strcmp(argv[arg], "create")) /* PVC */
      {
      fr_pvc(sock, &req, argv[++arg], 1);
      continue;
      }

    if (argc > arg+1 && !strcmp(argv[arg], "delete")) /* PVC */
      {
      fr_pvc(sock, &req, argv[++arg], -1);
      continue;
      }

    /* Download firmware and run board */
    if (argc > arg+1 && !strcmp(argv[arg], "run"))
      {
      run_board(sock, &req, argv[++arg]);
      continue;
      }
    
    /* Set mode */
    if (argc > arg+1 && !strcmp(argv[arg], "mode"))
      {
      int mode=0;
      arg++;

      if (argc > arg+1 && !strcmp(argv[arg], "soft"))
	{
	mode |= MODE_SOFT;
	arg++;
	}

      if (argc > arg+1 && !strcmp(argv[arg], "dce"))
	{
	mode |= MODE_DCE;
	arg++;
	}

      if (!strcmp(argv[arg], "hdlc"))
	mode |= MODE_HDLC;
      else if (!strcmp(argv[arg], "cisco"))
	mode |= MODE_CISCO;
      else if (!strcmp(argv[arg], "ppp"))
	mode |= MODE_PPP;
      else if (!strcmp(argv[arg], "x25"))
	mode |= MODE_X25;
      else if (!strcmp(argv[arg], "fr_ansi"))
	mode |= MODE_FR_ANSI;
      else if (!strcmp(argv[arg], "fr_ccitt"))
	mode |= MODE_FR_CCITT;
      else
	mode = 0;

      if (mode)
	{
	simple_ioctl(sock, &req, HDLCSETMODE, mode);
	continue;
	}
      }

    fprintf(stderr, "\nInvalid parameters\n");
    exit(1);
    }

  close(sock);
  exit(0);
  }
